# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os

from quixote import get_session, get_publisher, get_request, get_response
from quixote.directory import Directory, AccessControlled
from quixote.html import htmltext, TemplateIO

from qommon.admin.root import AdminDirectory

import settings
import forms
import roles
import users
import categories
import logger
import workflows
import bounces

from qommon import errors, get_cfg

def gpl():
    return htmltext("""<p>This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.</p>

    <p>This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.</p>

    <p>You should have received a copy of the GNU General Public License
    along with this program; if not, see <a href="http://www.gnu.org/licenses/"
    >http://www.gnu.org/licenses/</a>.</p>
    """)


class RootDirectory(AccessControlled, AdminDirectory):
    _q_exports = ['']

    settings = settings.SettingsDirectory()
    forms = forms.FormsDirectory()
    roles = roles.RolesDirectory()
    users = users.UsersDirectory()
    categories = categories.CategoriesDirectory()
    logger = logger.LoggerDirectory()
    workflows = workflows.WorkflowsDirectory()
    bounces = bounces.BouncesDirectory()

    menu_items = [
        ('forms/', N_('Forms')),
        ('workflows/', N_('Workflows')),
        ('users/', N_('Users')),
        ('roles/', N_('Roles')),
        ('categories/', N_('Categories')),
        ('logger/', N_('Logs')),
        ('bounces/', N_('Bounces')),
        ('settings/', N_('Settings')),
        ('/', N_('WCS Form Server'))]

    def _q_access(self):
        get_response().breadcrumb.append( ('admin/', _('Administration')) )

        req = get_request()

        if os.path.exists(os.path.join(get_publisher().app_dir, 'ADMIN_FOR_ALL')):
            get_response().filter['admin_for_all'] = True
            if req.user and req.user.is_admin:
                # if the user had access to the admin, ADMIN_FOR_ALL was
                # certainly added because something wrong happened when setting
                # fine-grained access permissions with roles; so we give the
                # user all possible roles.
                req.user.roles = [x.id for x in roles.Role.select()]
            return

        session = get_session()

        if req.user:
            if get_publisher().user_class.count() == 0:
                # this means user logged in anonymously
                pass
            elif not req.user.is_admin:
                raise errors.AccessForbiddenError()
        else:
            if get_publisher().user_class.count() > 0:
                raise errors.AccessUnauthorizedError()

        return


    def _q_index(self):
        from menu import html_top, get_vc_version
        html_top('/')
        r = TemplateIO(html=True)
        r+= htmltext('<p>%s</p>') % _('''
w.c.s. is a web application which allows to design and set up online forms.
It gives users the ability to create web forms easily without requiring any
other skill than familiarity with web surfing.''')
        r += htmltext('<p>%s</p>') % _('''w.c.s. is licensed under the terms of the GNU General
Public License.''')
        version = get_vc_version()
        if version:
            r += htmltext('<p class="version-info">')
            r += _('Version:')
            r += ' '
            r += version
            r += htmltext('</p>')

        r += htmltext('<p>%s</p>') % _('For more information:')

        r += htmltext('<ul>')
        r += htmltext('<li><a href="http://wcs.labs.libre-entreprise.org">%s</a></li>') % _('Web site')
        if get_publisher().admin_help_url and get_request().language in get_publisher().admin_help_url:
            doc_url = get_publisher().admin_help_url[get_request().language]
            r += htmltext('<li><a href="%s">%s</a></li>') % (doc_url, _('Online documentation'))
        r += htmltext('</ul>')

        get_response().filter['sidebar'] = str(self.get_sidebar())

        return r.getvalue()

    def get_sidebar(self):
        return gpl()

    def register_page(cls, url_name, directory = None, label = None):
        if directory:
            setattr(cls, url_name, directory)
        if not label:
            label = directory.label

        if directory and not url_name.endswith('/'):
            url_name += '/'

        if not (url_name, label) in cls.menu_items:
            logger_index = cls.menu_items.index(('logger/', 'Logs'))
            cls.menu_items.insert(logger_index, (url_name, label))

    register_page = classmethod(register_page)

    def _q_lookup(self, component):
        if not component in [str(x[0]).strip('/') for x in self.menu_items]:
            raise errors.TraversalError()

        authorised_roles = get_cfg('admin-permissions', {}).get(component)
        if authorised_roles and not os.path.exists(os.path.join(get_publisher().app_dir, 'ADMIN_FOR_ALL')):
            user_roles = set(get_request().user.roles)
            if not user_roles.intersection(authorised_roles):
                raise errors.AccessForbiddenError()

        return getattr(self, component)

    def get_menu_items(self):
        if not get_request().user:
            return self.menu_items
        user_roles = set(get_request().user.roles or [])
        if not get_cfg('admin-permissions', {}):
            return self.menu_items
        menu_items = self.menu_items[:]
        for k, v in self.menu_items:
            if not k.endswith(str('/')):
                continue
            k = k.strip(str('/'))
            if not k:
                continue
            authorised_roles = get_cfg('admin-permissions', {}).get(k)
            if not authorised_roles:
                continue
            if not user_roles.intersection(authorised_roles):
                menu_items.remove((k+'/', v))
        return menu_items
